// Copyright (c) The University of Edinburgh,  2007.
//See OGSA-DAI-Licence.txt for licencing information.

package org.biosphere.portlets.util;

import java.io.BufferedReader;
import java.io.Reader;
import java.net.URL;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.StringTokenizer;
//ogsadai-3.0-client.jar
import uk.org.ogsadai.client.toolkit.DataIterator;
import uk.org.ogsadai.client.toolkit.DataRequestExecutionResource;
import uk.org.ogsadai.client.toolkit.PipelineWorkflow;
import uk.org.ogsadai.client.toolkit.RequestExecutionType;
import uk.org.ogsadai.client.toolkit.RequestResource;
import uk.org.ogsadai.client.toolkit.ServerProxy;
import uk.org.ogsadai.client.toolkit.activities.delivery.DeliverToRequestStatus;
import uk.org.ogsadai.client.toolkit.activities.xmldb.XMLListCollections;
import uk.org.ogsadai.client.toolkit.activities.xmldb.XMLListResources;
import uk.org.ogsadai.client.toolkit.activities.xmldb.XPathQuery;
// import the base class for the example
import uk.org.ogsadai.client.toolkit.example.ClientBase;

// ogsadai-3.0-clientserver.jar 
import uk.org.ogsadai.resource.ResourceID;
import uk.org.ogsadai.resource.request.RequestStatus;

/**
 * Simple interface utility to XML DB via OGSA-DAI based on the XMDBClient example.
 * stuart@aiai.ed.ac.uk  28/4/2008
 * Simple command-line client for listing collections and
 * resources in XMLDB data resources and running XPath queries.
 * The client uses the client toolkit. 
 * <p>
 * <pre>
 * $ java uk.org.ogsadai.client.toolkit.example XMLDBClient
 *   -u SERVICES-BASE-URL - e.g. "http://localhost:8080/dai/services/";
 *   -e DRER-ID - e.g. DataRequestExecutionResource
 *   -d DATA-RESOURCE-ID - e.g. XMLDBResource
 *   -c listCollections - list collections.
 *   -coll COLLECTION - optional sub-collection to list.
 *
 * $ java uk.org.ogsadai.client.toolkit.example XMLDBClient
 *   -u SERVICES-BASE-URL - e.g. "http://localhost:8080/dai/services/";
 *   -e DRER-ID - e.g. DataRequestExecutionResource
 *   -d DATA-RESOURCE-ID - e.g. XMLDBResource
 *   -c listResources - list resources.
 *   -coll COLLECTION - optional sub-collection to list resources of.
 *
 * $ java uk.org.ogsadai.client.toolkit.example XMLDBClient
 *   -u  SERVICES-BASE-URL - e.g. "http://localhost:8080/dai/services/";
 *   -e  DRER-ID - e.g. DataRequestExecutionResource
 *   -d DATA-RESOURCE-ID - e.g. XMLDBResource
 *   -c xpath
 *   -coll COLLECTION - optional sub-collection to run query over.
 *   -r RESOURCE - optional resource to run query over.
 *   -n PREFIX:URN, PREFIX:URN,... - optional mappings from namespace
 *      prefixes used in the XPath query to namespace URNs.
 *   -x XPATH-QUERY - XPath query to run.
 * </pre>
 *
 * @author The OGSA-DAI Project Team
 */ 
public class XMLDBClientUtil extends ClientBase {
    /** Copyright statement */
    private static final String COPYRIGHT_NOTICE =
        "Copyright (c) The University of Edinburgh, 2007";

    /** List collections command. */
    private static final String LIST_COLLECTIONS = "listCollections";
    /** List resources command. */
    private static final String LIST_RESOURCES = "listResources";
    /** XPath query command. */
    private static final String XPATH_QUERY = "xpath";

    /** Command-line flag for URL. */
    private static final String URL= "u";
    /** Command-line flag for DRER. */
    private static final String DRER = "e";
    /** Command-line flag for data resource. */
    private static final String DATA_RESOURCE = "d";
    /** Command-line flag for command. */
    private static final String COMMAND = "c";
    /** Command-line flag for collection. */
    private static final String COLLECTION = "coll";
    /** Command-line flag for resource. */
    private static final String RESOURCE = "r";
    /** Command-line flag for XPath query. */
    private static final String XPATH = "x";
    /** Command-line flag for namespace mappings. */
    private static final String NAMESPACE = "n";

    static StringBuffer result;

    /**
     * Run from command line.
     * 
     * @param args 
     *     Command-line arguments.
     * @throws Exception
     *     If there are any problems.
     */
    public XMLDBClientUtil()
    {
	this.result = new StringBuffer();
    }

    /**
     * Parse the arguments, submit a request to the OGSA-DAI server
     * and pretty-print the results.
     * 
     * @param args 
     *     Command-line arguments.
     *     
     * @throws Exception if an error occurs.
     */
    public String execute(String args_string) throws Exception
    {
	// convert the argument string into a commandline String[]
	Vector args_vec = new Vector();
	String next;
	StringTokenizer strtoken = new StringTokenizer(args_string," \n",false);
	for(;strtoken.hasMoreElements();) 
	    args_vec.addElement(strtoken.nextToken());
	String[] args = new String[args_vec.size()];
	for(int i=0;i<args_vec.size();++i)
	    args[i]=(String)args_vec.elementAt(i);
	return execute(args);
    }

    // back to the original code
    public String execute(String[] args) throws Exception
    {
        Map parameters = new Hashtable();
        super.parseArguments(args, parameters);

        // Initialise arguments.
        String url = super.getSingleValueOptionalFlag(parameters, URL);
        if (url == null)
        {
            url = "http://localhost:8080/dai/services/";
        }
        String drerIDStr = super.getSingleValueOptionalFlag(parameters, DRER);
        if (drerIDStr == null)
        {
            drerIDStr = "DataRequestExecutionResource";
        }
        ResourceID drerID = new ResourceID(drerIDStr);
        String dataResourceIDStr = 
            super.getSingleValueOptionalFlag(parameters, DATA_RESOURCE);
        if (dataResourceIDStr == null)
        {
            dataResourceIDStr = "XMLDBResource";
        }
        ResourceID dataResourceID = new ResourceID(dataResourceIDStr);
        String command = super.getSingleValue(parameters, COMMAND);

        String collection = null;
        String resource = null;
        String xpath = null;
        List namespaces = null;

        if (command.equals(LIST_COLLECTIONS))
        {
            collection = getSingleValueOptionalFlag(parameters, COLLECTION);
        }
        else if (command.equals(LIST_RESOURCES))
        {
            collection = getSingleValueOptionalFlag(parameters, COLLECTION);
        }
        else if (command.equals(XPATH_QUERY))
        {
            collection = getSingleValueOptionalFlag(parameters, COLLECTION);
            resource = getSingleValueOptionalFlag(parameters, RESOURCE);
            namespaces = getValuesOptionalFlag(parameters, NAMESPACE);
            xpath = getSingleValue(parameters, XPATH);
        }
        else
        {
            throw new IllegalArgumentException("value for flag -" + COMMAND + 
                                               " must be one of: " + 
                                               LIST_COLLECTIONS + ", " +
                                               LIST_RESOURCES + " or " +
                                               XPATH_QUERY);
        }

        // Print.
        append_nl("DRER ID: " + drerID);
        append_nl("Data Resource ID: " + dataResourceID);
        append_nl("Base Services URL: " + url);
        append_nl("Command: " + command);
        if (collection != null)
        {
            append_nl("Collection: " + collection);
        }
        if (resource != null)
        {
            append_nl("Resource: " + resource);
        }
        if (xpath != null)
        {
            append_nl("XPath: " + xpath);
        }
        // Print.
        append_nl("DRER ID: " + drerID);
        append_nl("Data Resource ID: " + dataResourceID);
        append_nl("Base Services URL: " + url);
        append_nl("Command: " + command);
        if (collection != null)
        {
            append_nl("Collection: " + collection);
        }
        if (resource != null)
        {
            append_nl("Resource: " + resource);
        }
        if (xpath != null)
        {
            append_nl("XPath: " + xpath);
        }

        // Get server.
        ServerProxy server = new ServerProxy();
        server.setDefaultBaseServicesURL(new URL(url));

        // Get DRER
        DataRequestExecutionResource drer = 
            server.getDataRequestExecutionResource(drerID);

        if (command.equals(LIST_COLLECTIONS))
        {
            listCollections(drer, dataResourceID, collection);
        }
        else if (command.equals(LIST_RESOURCES))
        {
            listResources(drer, dataResourceID, collection);
        }
        else if (command.equals(XPATH_QUERY))
        {
            xpathQuery(drer, dataResourceID, collection, resource, 
                       namespaces, xpath);  
        }
	return this.result.toString();
    }

    /**
     * Submit a request to a DRER to list the collections of
     * an XMLDB data resource.
     *
     * @param drer
     *      Client toolkit data request execution resource proxy.
     * @param dataResourceID
     *      ID of XMLDB data resource.
     * @param collection
     *      Sub-collection of XMLDB data resource whose collections are
     *      to be listed. If <code>null</code> then the the collections
     *      of the XMLDB data resource base collection are listed.
     * @throws Exception
     *     If there are any problems.
     */
    private static void listCollections(DataRequestExecutionResource drer,
                                        ResourceID dataResourceID,
                                        String collection)
        throws Exception
    {
        // List collections.
        XMLListCollections listCollections = new XMLListCollections();
        listCollections.setResourceID(dataResourceID);
        if (collection != null)
        {
            listCollections.addParentCollection(collection);
        }
        // Deliver to request status.
        DeliverToRequestStatus deliverToRequestStatus = 
            new DeliverToRequestStatus();
        deliverToRequestStatus.connectInput(listCollections.getDataOutput());

        // Workflow.
        PipelineWorkflow pipeline = new PipelineWorkflow();
        pipeline.add(listCollections);
        pipeline.add(deliverToRequestStatus);

        // Execute.
        RequestResource requestResource = null;
        try
        {
            requestResource = 
                drer.execute(pipeline, RequestExecutionType.SYNCHRONOUS);
        }
        catch (Throwable e)
        {
            append_nl("A problem has occured...listCollections()");
            while (e != null)
            {
                append_nl(e.getMessage());
                e = e.getCause();
            }
            return; // SA System.exit(1);
        }
        RequestStatus status = requestResource.getRequestStatus();
        // Get request execution status.
        append_nl(status.getExecutionStatus());

        // Get results.
        while (listCollections.hasNextData())
        {
            DataIterator iterator = listCollections.nextData();
            while (iterator.hasNext())
            {
                append_nl(iterator.next());
            }
        }
    }

    /**
     * Submit a request to a DRER to list the resources of
     * an XMLDB data resource.
     *
     * @param drer
     *      Client toolkit data request execution resource proxy.
     * @param dataResourceID
     *      ID of XMLDB data resource.
     * @param collection
     *      Sub-collection of XMLDB data resource whose resources are
     *      to be listed. If <code>null</code> then the the resource
     *      of the XMLDB data resource base collection are listed.
     * @throws Exception
     *     If there are any problems.
     */
    private static void listResources(DataRequestExecutionResource drer,
                                      ResourceID dataResourceID,
                                      String collection)
        throws Exception
    {
        // Build the individual activities making the workflow
        // and connect them up.
        XMLListResources listResources = new XMLListResources();
        listResources.setResourceID(dataResourceID);
        if (collection != null)
        {
            listResources.addParentCollection(collection);
        }
        // Deliver to request status.
        DeliverToRequestStatus deliverToRequestStatus = 
            new DeliverToRequestStatus();
        deliverToRequestStatus.connectInput(listResources.getDataOutput());

        // Workflow.
        PipelineWorkflow pipeline = new PipelineWorkflow();
        pipeline.add(listResources);
        pipeline.add(deliverToRequestStatus);

        // Execute.
        RequestResource requestResource = null;
        try
        {
            requestResource = 
                drer.execute(pipeline, RequestExecutionType.SYNCHRONOUS);
        }
        catch (Throwable e)
        {
            append_nl("A problem has occured...listResources()");
            while (e != null)
            {
                append_nl(e.getMessage());
                e = e.getCause();
            }
            return; // SA System.exit(1);
        }
        RequestStatus status = requestResource.getRequestStatus();
        // Get request execution status.
        append_nl(status.getExecutionStatus());

        // Get results.
        while (listResources.hasNextData())
        {
            DataIterator iterator = listResources.nextData();
            while (iterator.hasNext())
            {
                append_nl(iterator.next());
            }
        }
    }

    /**
     * Submit a request to a DRER to run an XPath query over the resources 
     * of an XMLDB data resource.
     *
     * @param drer
     *      Client toolkit data request execution resource proxy.
     * @param dataResourceID
     *      ID of XMLDB data resource.
     * @param collection
     *      Sub-collection of XMLDB data resource whose resources are
     *      to be queried. If <code>null</code> then the collection is
     *      is assumed to be the XMLDB base collection.
     * @param resource
     *      Specific resource to be queried. If <code>null</code> then
     *      all the resources of the collection specified by the above
     *      are queried.
     * @param namespaces
     *      Namespace mappings as strings of form <code>PREFIX:URN</code>.
     * @param xpath
     *     XPath query expression.
     * @throws Exception
     *     If there are any problems.
     */
    private static void xpathQuery(DataRequestExecutionResource drer,
                                   ResourceID dataResourceID,
                                   String collection,
                                   String resource,
                                   List namespaces,
                                   String xpath)
        throws Exception
    {
        // XPath.
        XPathQuery xpathStatement = new XPathQuery();
        xpathStatement.setResourceID(dataResourceID);
        if (collection != null)
        {
            xpathStatement.addCollection(collection);
        }
        if (resource != null)
        {
            xpathStatement.addResourceID(resource);
        }
        xpathStatement.addExpression(xpath);
        if (namespaces != null)
        {
            xpathStatement.addNamespace(namespaces.iterator());
        }
        // Deliver to request status.
        DeliverToRequestStatus deliverToRequestStatus = 
            new DeliverToRequestStatus();
        deliverToRequestStatus.connectInput(xpathStatement.getDataOutput());

        // Workflow.
        PipelineWorkflow pipeline = new PipelineWorkflow();
        pipeline.add(xpathStatement);
        pipeline.add(deliverToRequestStatus);

        // Execute.
        RequestResource requestResource = null;
        try
        {
            requestResource = 
                drer.execute(pipeline, RequestExecutionType.SYNCHRONOUS);
        }
        catch (Throwable e)
        {
            append_nl("A problem has occured...");
            while (e != null)
            {
                append_nl(e.getMessage());
                e = e.getCause();
            }
            return; // SA System.exit(1);
        }
        RequestStatus status = requestResource.getRequestStatus();
        // Get request execution status.
        append_nl(status.getExecutionStatus());

        // Get results.
        while (xpathStatement.hasNextResult())
        {
            Reader reader = xpathStatement.getNextResult();
            BufferedReader buffReader = new BufferedReader(reader);
            String line = null;
            while ((line = buffReader.readLine()) != null)
            {
                append_nl(line);
            }
        }
    }

    public static void append_nl(Object o) {
	result.append(o.toString()+"<br>\n");
    }

    public static void main(String[] args) {
	if(args.length==0) {
	    System.out.println("Usage: set classpath correctly, ensure tomcat is running, and OGSA-DAI and eXist are deployed \nrun at command line: java org.biosphere.portlets.util.XMLDBClientUtil   -u http://localhostl:8080/dai/services -d ExistDataResource -c listCollections");
	    XMLDBClientUtil client = new XMLDBClientUtil();
	    String result = "";
	    try {
		result = client.execute("-u http://localhostl:8080/dai/services -d ExistDataResource -c listCollections");
	    }
	    catch(Exception e) {
		System.out.println("error: exception calling execute");
	    }
	    System.out.println(result);
	}
	else {
	    XMLDBClientUtil client = new XMLDBClientUtil();
	    String result = "";
	    try {
		result = client.execute(args);
	    }
	    catch(Exception e) {
		System.out.println("error: exception calling execute");
	    }
	    System.out.println(result);
	}
    }
}
